/*
 Java Class Inspector
   Copyright (C) 1996 Clark VERBRUGGE.     Contact: clump@cs.mcgill.ca
   Copyright (C) 1996 McGill University. 
   Copyright (C) 1996 McCAT System Group.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 1, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 675
   Mass Ave, Cambridge, MA 02139, USA. */

import java.io.*;

public class ClassFile {

	// Magic number of classfile
	public static final int MAGIC = 0xCAFEBABE;

	// Access flags
	public static final short ACC_PUBLIC =    0x0001;
	public static final short ACC_PRIVATE =   0x0002;
	public static final short ACC_PROTECTED = 0x0004;
	public static final short ACC_STATIC =    0x0008;
	public static final short ACC_FINAL =     0x0010;
	public static final short ACC_SUPER =     0x0020;
	public static final short ACC_VOLATILE =  0x0040;
	public static final short ACC_TRANSIENT = 0x0080;
	public static final short ACC_INTERFACE = 0x0200;
	public static final short ACC_ABSTRACT =  0x0400;

	public static final short ACC_UNKNOWN =   0x7800; // whatever's left

	// descriptor strings
	public static final String DESC_BYTE =    "B";
	public static final String DESC_CHAR =    "C";
	public static final String DESC_DOUBLE =  "D";
	public static final String DESC_FLOAT=    "F";
	public static final String DESC_INT =     "I";
	public static final String DESC_LONG =    "J";
	public static final String DESC_OBJECT =  "L";
	public static final String DESC_SHORT =   "S";
	public static final String DESC_BOOLEAN = "Z";
	public static final String DESC_VOID =    "V";
	public static final String DESC_ARRAY =   "[";

	// values for inserted profiling code
	public static final String HASHTABLEVAR = "___PROFILE_HASHTABLE_";
	public static final String HASHTABLE = "java/util/Hashtable";
	public static final String HASHTABLEDESC = DESC_OBJECT + HASHTABLE + ";";
	public static final String HASHTABLEPUT = "put";
	public static final String HASHTABLEGET = "get";
	public static final String OBJECT = "java/lang/Object";
	public static final String OBJECTDESC = DESC_OBJECT + OBJECT + ";";

	public boolean debug;

    String fn;

	public int magic;
	public short minor_version;
	public short major_version;
	public short constant_pool_count;
	public cp_info constant_pool[];
	public short access_flags;
	public short this_class;
	public short super_class;
	public short interfaces_count;
	public short interfaces[];
	public short fields_count;
	public field_info fields[];
	public short methods_count;
	public method_info methods[];
	public short attributes_count;
	public attribute_info attributes[];

    public ClassFile(String nfn) { fn = nfn; }

	// Returns a string with the appropriate strings representing each
	// bit of af.
	public String access_string(short af,String separator) {
		boolean hasone = false;
		String s = "";
		if ((af & ACC_PUBLIC) != 0) {
			s = "PUBLIC";
			hasone = true;
		}
		if ((af & ACC_PRIVATE) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "PRIVATE";
		}
		if ((af & ACC_PROTECTED) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "PROTECTED";
		}
		if ((af & ACC_STATIC) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "STATIC";
		}
		if ((af & ACC_FINAL) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "FINAL";
		}
		if ((af & ACC_SUPER) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "SUPER";
		}
		if ((af & ACC_VOLATILE) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "VOLATILE";
		}
		if ((af & ACC_TRANSIENT) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "TRANSIENT";
		}
		if ((af & ACC_INTERFACE) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "INTERFACE";
		}
		if ((af & ACC_ABSTRACT) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "ABSTRACT";
		}
		if ((af & ACC_UNKNOWN) != 0) {
			if (hasone) s = s + separator;
			else hasone = true;
			s = s + "UNKNOWN";
		}
		return s;
	}

	public boolean read(DataInputStream d) {
		try {
			// first read in magic number
			magic = d.readInt();
			if (magic != MAGIC) {
				System.out.println("Wrong magic number in " + fn + ": " + magic);
				return false;
			}
			System.out.println("Magic number ok");
			minor_version = (short)d.readUnsignedShort();
			major_version = (short)d.readUnsignedShort();
			System.out.println("Version: " + major_version + "." + minor_version);
			constant_pool_count = (short)d.readUnsignedShort();
			System.out.println("Constant pool count: " + constant_pool_count);

			if (!readConstantPool(d))
				return false;

			access_flags = (short)d.readUnsignedShort();
			System.out.println("Access flags: " + access_flags + " = " + 
			                   access_string(access_flags,", "));

			this_class = (short)d.readUnsignedShort();
			super_class = (short)d.readUnsignedShort();
			interfaces_count = (short)d.readUnsignedShort();
			if (interfaces_count>0) {
				interfaces = new short[interfaces_count];
				int j;
				for (j=0; j<interfaces_count; j++) 
					interfaces[j] = (short)d.readUnsignedShort();
			}
			System.out.println("Implements " + interfaces_count + " interface(s)");

			fields_count = (short)d.readUnsignedShort();
			System.out.println("Has " + fields_count + " field(s)");
			readFields(d);

			methods_count = (short)d.readUnsignedShort();
			System.out.println("Has " + methods_count + " method(s)");
			readMethods(d);

			attributes_count = (short)d.readUnsignedShort();
			System.out.println("Has " + attributes_count + " attribute(s)");
			if (attributes_count>0) {
				attributes =  new attribute_info[attributes_count];
				readAttributes(d,attributes_count,attributes);
			}
		} catch(IOException e) {
			System.out.println("IOException with " + fn + ": " + e.getMessage());
			return false;
		}
		return true;
	}

	protected boolean readConstantPool(DataInputStream d) throws IOException {
		byte tag;
		cp_info cp;
		int i;
		boolean skipone;   // set if next cp entry is to be skipped

		constant_pool = new cp_info[constant_pool_count];
		Instruction.constant_pool = constant_pool;
		skipone = false;

		for (i=1;i<constant_pool_count;i++) {
			if (skipone) {
				skipone = false;
				continue;
			}
			tag = (byte)d.readUnsignedByte();
			switch(tag) {
				case cp_info.CONSTANT_Class:
				  cp = new CONSTANT_Class_info();
				  ((CONSTANT_Class_info)cp).name_index = (short)d.readUnsignedShort();
				  if (debug) System.out.println("Constant pool[" + i + "]: Class");
				  break;
				case cp_info.CONSTANT_Fieldref:
				  cp = new CONSTANT_Fieldref_info();
				  ((CONSTANT_Fieldref_info)cp).class_index = (short)d.readUnsignedShort();
				  ((CONSTANT_Fieldref_info)cp).name_and_type_index = (short)d.readUnsignedShort();
				  if (debug) System.out.println("Constant pool[" + i + "]: Fieldref");
				  break;
				case cp_info.CONSTANT_Methodref:
				  cp = new CONSTANT_Methodref_info();
				  ((CONSTANT_Methodref_info)cp).class_index = (short)d.readUnsignedShort();
				  ((CONSTANT_Methodref_info)cp).name_and_type_index = (short)d.readUnsignedShort();
				  if (debug) System.out.println("Constant pool[" + i + "]: Methodref");
				  break;
				case cp_info.CONSTANT_InterfaceMethodref:
				  cp = new CONSTANT_InterfaceMethodref_info();
				  ((CONSTANT_InterfaceMethodref_info)cp).class_index = 
				  	(short)d.readUnsignedShort();
				  ((CONSTANT_InterfaceMethodref_info)cp).name_and_type_index = 
				  	(short)d.readUnsignedShort();
				  if (debug) System.out.println("Constant pool[" + i + "]: InterfaceMethodref");
				  break;
				case cp_info.CONSTANT_String:
				  cp = new CONSTANT_String_info();
				  ((CONSTANT_String_info)cp).string_index = 
				  	(short)d.readUnsignedShort();
				  if (debug) System.out.println("Constant pool[" + i + "]: String");
				  break;
				case cp_info.CONSTANT_Integer:
				  cp = new CONSTANT_Integer_info();
				  ((CONSTANT_Integer_info)cp).bytes = d.readInt();
				  if (debug) System.out.println("Constant pool[" + i + "]: Integer = " + 
				                                ((CONSTANT_Integer_info)cp).bytes);
				  break;
				case cp_info.CONSTANT_Float:
				  cp = new CONSTANT_Float_info();
				  ((CONSTANT_Float_info)cp).bytes = d.readInt();
				  if (debug) System.out.println("Constant pool[" + i + "]: Float = " + 
				                                ((CONSTANT_Float_info)cp).convert());
				  break;
				case cp_info.CONSTANT_Long:
				  cp = new CONSTANT_Long_info();
				  ((CONSTANT_Long_info)cp).high = d.readInt();
				  ((CONSTANT_Long_info)cp).low = d.readInt();
				  if (debug) {
				     String temp = cp.toString(constant_pool);
				     System.out.println("Constant pool[" + i + "]: Long = " + temp);
				     /*System.out.println("Constant pool[" + i + "]: that's " +
				                        cp.printBits(((CONSTANT_Long_info)cp).high) + " <<32 + " + 
				                        cp.printBits(((CONSTANT_Long_info)cp).low) + " = " + 
				                        cp.printBits(((CONSTANT_Long_info)cp).convert()));*/
				  }
				  skipone = true;  // next entry needs to be skipped
				  break;
				case cp_info.CONSTANT_Double:
				  cp = new CONSTANT_Double_info();
				  ((CONSTANT_Double_info)cp).high = d.readInt();
				  ((CONSTANT_Double_info)cp).low = d.readInt();
				  if (debug) System.out.println("Constant pool[" + i + "]: Double = " + 
				                                ((CONSTANT_Double_info)cp).convert());
				  skipone = true;  // next entry needs to be skipped
				  break;
				case cp_info.CONSTANT_NameAndType:
				  cp = new CONSTANT_NameAndType_info();
				  ((CONSTANT_NameAndType_info)cp).name_index = (short)d.readUnsignedShort();
				  ((CONSTANT_NameAndType_info)cp).descriptor_index = (short)d.readUnsignedShort();
				  if (debug) System.out.println("Constant pool[" + i + "]: Name and Type");
				  break;
				case cp_info.CONSTANT_Utf8:
				  short len;
				  CONSTANT_Utf8_info cputf8 = new CONSTANT_Utf8_info();
				  len = (short)d.readUnsignedShort();
				  cputf8.bytes = new byte[len+2];
				  cputf8.bytes[0] = (byte)(len>>8);
				  cputf8.bytes[1] = (byte)(len & 0xff);
				  if (len>0) {
				  	  int j;
				  	  for (j=0; j<len;j++)
				    	cputf8.bytes[j+2] = (byte)d.readUnsignedByte();				  	
				  }
				  cp = (cp_info)cputf8;
				  if (debug) System.out.println("Constant pool[" + i + "]: Utf8 = \""
				                                + cputf8.convert() + "\"");
				  break;
				default: 
				  System.out.println("Unknown tag in constant pool: " + tag + " at entry " + i);
				  return false;
			}
			cp.tag = tag;
			constant_pool[i] = cp;
		}
		return true;
	}

	protected boolean readAttributes(DataInputStream d,short attributes_count,attribute_info[] ai) throws IOException {
		attribute_info a=null;
		int i,len;
		short j;
		String s;

		for (i=0;i<attributes_count;i++) {
			j = (short)d.readUnsignedShort();  // read attribute name before allocating
			len = d.readInt();
			s = ((CONSTANT_Utf8_info)(constant_pool[j])).convert();

			if (s.compareTo(attribute_info.SourceFile)==0) {
				SourceFile_attribute sa = new SourceFile_attribute();
				sa.sourcefile_index = (short)d.readUnsignedShort();
				a = (attribute_info)sa;
			} else if(s.compareTo(attribute_info.ConstantValue)==0) {
				ConstantValue_attribute ca = new ConstantValue_attribute();
				ca.constantvalue_index = (short)d.readUnsignedShort();
				a = (attribute_info)ca;
			} else if(s.compareTo(attribute_info.Code)==0) {
				Code_attribute ca = new Code_attribute();
				ca.max_stack = (short)d.readUnsignedShort();
				ca.max_locals = (short)d.readUnsignedShort();
				ca.code_length = d.readInt();
				ca.code = new byte[ca.code_length];
				d.read(ca.code);
				ca.exception_table_length = (short)d.readUnsignedShort();
				ca.exception_table = new exception_table_entry[ca.exception_table_length];
				int k;
				exception_table_entry e;
				for (k=0; k<ca.exception_table_length; k++) {
					e = new exception_table_entry();
					e.start_pc = (short)d.readUnsignedShort();
					e.end_pc = (short)d.readUnsignedShort();
					e.handler_pc = (short)d.readUnsignedShort();
					e.catch_type = (short)d.readUnsignedShort();
					ca.exception_table[k] = e;
				}
				ca.attributes_count = (short)d.readUnsignedShort();
				ca.attributes = new attribute_info[ca.attributes_count];
				readAttributes(d,ca.attributes_count,ca.attributes);
				a = (attribute_info)ca;
			} else if(s.compareTo(attribute_info.Exceptions)==0) {
				Exception_attribute ea = new Exception_attribute();
				ea.number_of_exceptions = (short)d.readUnsignedShort();
				if (ea.number_of_exceptions>0) {
					int k;
					ea.exception_index_table = new short[ea.number_of_exceptions];
					for (k=0; k<ea.number_of_exceptions; k++) 
						ea.exception_index_table[k]  = (short)d.readUnsignedShort();
				}
				a = (attribute_info)ea;
			} else if(s.compareTo(attribute_info.LineNumberTable)==0) {
				LineNumberTable_attribute la = new LineNumberTable_attribute();
				la.line_number_table_length = (short)d.readUnsignedShort();
				int k;
				line_number_table_entry e;
				la.line_number_table = new line_number_table_entry[la.line_number_table_length];
				for (k=0; k<la.line_number_table_length; k++) {
					e = new line_number_table_entry();
					e.start_pc = (short)d.readUnsignedShort();
					e.line_number = (short)d.readUnsignedShort();
					la.line_number_table[k] = e;
				}
				a = (attribute_info)la;
			} else if(s.compareTo(attribute_info.LocalVariableTable)==0) {
				LocalVariableTable_attribute la = new LocalVariableTable_attribute();
				la.local_variable_table_length = (short)d.readUnsignedShort();
				int k;
				local_variable_table_entry e;
				la.local_variable_table = 
					new local_variable_table_entry[la.local_variable_table_length];
				for (k=0; k<la.local_variable_table_length; k++) {
					e = new local_variable_table_entry();
					e.start_pc = (short)d.readUnsignedShort();
					e.length = (short)d.readUnsignedShort();
					e.name_index = (short)d.readUnsignedShort();
					e.descriptor_index = (short)d.readUnsignedShort();
					e.index = (short)d.readUnsignedShort();
					la.local_variable_table[k] = e;
				}
				a = (attribute_info)la;
			} else {
				// unknown attribute
				System.out.println("Generic/Unknown Attribute: " + s);
				Generic_attribute ga = new Generic_attribute();
				if (len>0) {
					ga.info = new byte[len];
					d.read(ga.info);
				}
				a = (attribute_info)ga;
			}
			a.attribute_name = j;		
			a.attribute_length = len;
			ai[i] = a;
		}
		return true;
	}


	protected boolean readFields(DataInputStream d) throws IOException {
		field_info fi;
		int i;

		fields = new field_info[fields_count];

		for (i=0;i<fields_count;i++) {
			fi = new field_info();
			fi.access_flags = (short)d.readUnsignedShort();
			fi.name_index = (short)d.readUnsignedShort();
			fi.descriptor_index = (short)d.readUnsignedShort();
			fi.attributes_count = (short)d.readUnsignedShort();
			if (fi.attributes_count>0) {
				fi.attributes = new attribute_info[fi.attributes_count];
				readAttributes(d,fi.attributes_count,fi.attributes);
			}
			/*CONSTANT_Utf8_info ci;
			ci = (CONSTANT_Utf8_info)(constant_pool[fi.name_index]);
			System.out.println("Field: " + ci.convert());*/
			fields[i] = fi;
		}

		return true;
	}

	protected boolean readMethods(DataInputStream d) throws IOException {
		method_info mi;
		int i;

		methods = new method_info[methods_count];

		for (i=0;i<methods_count;i++) {
			mi = new method_info();
			mi.access_flags = (short)d.readUnsignedShort();
			mi.name_index = (short)d.readUnsignedShort();
			mi.descriptor_index = (short)d.readUnsignedShort();
			mi.attributes_count = (short)d.readUnsignedShort();

			/*CONSTANT_Utf8_info ci;
			ci = (CONSTANT_Utf8_info)(constant_pool[mi.name_index]);
			System.out.println(" " + access_string(mi.access_flags," ").toLowerCase() + 
			                   " " + ci.convert());*/

			if (mi.attributes_count>0) {
				mi.attributes = new attribute_info[mi.attributes_count];
				readAttributes(d,mi.attributes_count,mi.attributes);
			}

			/*if ("main".compareTo(ci.convert())==0) {
				decompile(mi);
			}*/

			methods[i] = mi;
		}

		return true;
	}

	public void showByteCode(Code_attribute ca) {
		int i=0,j;
		
		System.out.println("Code bytes follow...");
		while(i<ca.code_length) {
			j = (int)(ca.code[i]);
			j &= 0xff;
			System.out.print(Integer.toString(j) + " ");
			i++;
		}
		System.out.println("");
	}

	// returns the code attribute for method m
	public static Code_attribute locate_code_attribute(method_info m) {
		attribute_info ai;
		int i;

		for (i=0; i<m.attributes_count; i++) {
			ai = m.attributes[i];
			if (ai instanceof Code_attribute)
				return (Code_attribute)ai;
		}
		return null;
	}

	protected boolean writeConstantPool(DataOutputStream dd) throws IOException {
		byte tag;
		cp_info cp;
		int i;
		boolean skipone = false;

		for (i=1;i<constant_pool_count;i++) {
			if (skipone) {
				skipone = false;
				continue;
			}
			cp = constant_pool[i];
			dd.writeByte(cp.tag);
			switch(cp.tag) {
				case cp_info.CONSTANT_Class:
				  dd.writeShort(((CONSTANT_Class_info)cp).name_index);
				  break;
				case cp_info.CONSTANT_Fieldref:
				  dd.writeShort(((CONSTANT_Fieldref_info)cp).class_index);
				  dd.writeShort(((CONSTANT_Fieldref_info)cp).name_and_type_index);
				  break;
				case cp_info.CONSTANT_Methodref:
				  dd.writeShort(((CONSTANT_Methodref_info)cp).class_index);
				  dd.writeShort(((CONSTANT_Methodref_info)cp).name_and_type_index);
				  break;
				case cp_info.CONSTANT_InterfaceMethodref:
				  dd.writeShort(((CONSTANT_InterfaceMethodref_info)cp).class_index);
				  dd.writeShort(((CONSTANT_InterfaceMethodref_info)cp).name_and_type_index);
				  break;
				case cp_info.CONSTANT_String:
				  dd.writeShort(((CONSTANT_String_info)cp).string_index);
				  break;
				case cp_info.CONSTANT_Integer:
				  dd.writeInt(((CONSTANT_Integer_info)cp).bytes);
				  break;
				case cp_info.CONSTANT_Float:
				  dd.writeInt(((CONSTANT_Float_info)cp).bytes);
				  break;
				case cp_info.CONSTANT_Long:
				  dd.writeInt(((CONSTANT_Long_info)cp).high);
				  dd.writeInt(((CONSTANT_Long_info)cp).low);
				  skipone = true;
				  break;
				case cp_info.CONSTANT_Double:
				  dd.writeInt(((CONSTANT_Double_info)cp).high);
				  dd.writeInt(((CONSTANT_Double_info)cp).low);
				  skipone = true;
				  break;
				case cp_info.CONSTANT_NameAndType:
				  dd.writeShort(((CONSTANT_NameAndType_info)cp).name_index);
				  dd.writeShort(((CONSTANT_NameAndType_info)cp).descriptor_index);
				  break;
				case cp_info.CONSTANT_Utf8:
				  int len;
				  len = ((CONSTANT_Utf8_info)cp).bytes.length;
				  dd.writeShort(len-2);
				  dd.write(((CONSTANT_Utf8_info)cp).bytes,2,len-2);
				  break;
				default: 
				  System.out.println("Unknown tag in constant pool: " + cp.tag);
				  return false;
			}
		}
		return true;
	}

	protected boolean writeAttributes(DataOutputStream dd,short attributes_count,attribute_info[] ai) throws IOException {
		attribute_info a=null;
		int i,len;
		short j;
		String s;

		for (i=0;i<attributes_count;i++) {
			a = ai[i];
			dd.writeShort(a.attribute_name);
			dd.writeInt(a.attribute_length);
			if (a instanceof SourceFile_attribute) {
				SourceFile_attribute sa = (SourceFile_attribute)a;
				dd.writeShort(sa.sourcefile_index);
			} else if(a instanceof ConstantValue_attribute) {
				ConstantValue_attribute ca = (ConstantValue_attribute)a;
				dd.writeShort(ca.constantvalue_index);
			} else if(a instanceof Code_attribute) {
				Code_attribute ca = (Code_attribute)a;
				dd.writeShort(ca.max_stack);
				dd.writeShort(ca.max_locals);
				dd.writeInt(ca.code_length);
				dd.write(ca.code,0,ca.code_length);
				dd.writeShort(ca.exception_table_length);
				int k;
				exception_table_entry e;
				for (k=0; k<ca.exception_table_length; k++) {
					e = ca.exception_table[k];
					dd.writeShort(e.start_pc);
					dd.writeShort(e.end_pc);
					dd.writeShort(e.handler_pc);
					dd.writeShort(e.catch_type);
				}
				dd.writeShort(ca.attributes_count);
				if (ca.attributes_count>0)
					writeAttributes(dd,ca.attributes_count,ca.attributes);
			} else if(a instanceof Exception_attribute) {
				Exception_attribute ea = (Exception_attribute)a;
				dd.writeShort(ea.number_of_exceptions);
				if (ea.number_of_exceptions>0) {
					int k;
					for (k=0; k<ea.number_of_exceptions; k++) 
						dd.writeShort(ea.exception_index_table[k]);
				}
			} else if(a instanceof LineNumberTable_attribute) {
				LineNumberTable_attribute la = (LineNumberTable_attribute)a;
				dd.writeShort(la.line_number_table_length);
				int k;
				line_number_table_entry e;
				for (k=0; k<la.line_number_table_length; k++) {
					e = la.line_number_table[k];
					dd.writeShort(e.start_pc);
					dd.writeShort(e.line_number);
				}
			} else if(a instanceof LocalVariableTable_attribute) {
				LocalVariableTable_attribute la = (LocalVariableTable_attribute)a;
				dd.writeShort(la.local_variable_table_length);
				int k;
				local_variable_table_entry e;
				for (k=0; k<la.local_variable_table_length; k++) {
					e = la.local_variable_table[k];
					dd.writeShort(e.start_pc);
					dd.writeShort(e.length);
					dd.writeShort(e.name_index);
					dd.writeShort(e.descriptor_index);
					dd.writeShort(e.index);
				}
			} else {
				// unknown attribute
				System.out.println("Generic/Unknown Attribute in output");
				Generic_attribute ga = (Generic_attribute)a;
				if (ga.attribute_length>0) {
					dd.write(ga.info,0,ga.attribute_length);
				}
			}
		}
		return true;
	}

	protected boolean writeFields(DataOutputStream dd) throws IOException {
		field_info fi;
		int i;

		for (i=0;i<fields_count;i++) {
			fi = fields[i];
			dd.writeShort(fi.access_flags);
			dd.writeShort(fi.name_index);
			dd.writeShort(fi.descriptor_index);
			dd.writeShort(fi.attributes_count);
			if (fi.attributes_count>0) {
				writeAttributes(dd,fi.attributes_count,fi.attributes);
			}
		}
		return true;
	}

	protected boolean writeMethods(DataOutputStream dd) throws IOException {
		method_info mi;
		int i;

		for (i=0;i<methods_count;i++) {
			mi = methods[i];
			dd.writeShort(mi.access_flags);
			dd.writeShort(mi.name_index);
			dd.writeShort(mi.descriptor_index);
			dd.writeShort(mi.attributes_count);
			if (mi.attributes_count>0) {
				writeAttributes(dd,mi.attributes_count,mi.attributes);
			}
		}
		return true;
	}

	public boolean writeClass(DataOutputStream dd) {
		// outputs the .class file from the loaded one
		try {
			// first write magic number
			dd.writeInt(magic);

			dd.writeShort(minor_version);
			dd.writeShort(major_version);
			dd.writeShort(constant_pool_count);

			if (!writeConstantPool(dd))
				return false;

			dd.writeShort(access_flags);
			dd.writeShort(this_class);
			dd.writeShort(super_class);
			dd.writeShort(interfaces_count);
			if (interfaces_count>0) {
				int j;
				for (j=0; j<interfaces_count; j++) 
					dd.writeShort(interfaces[j]);
			}

			dd.writeShort(fields_count);
			writeFields(dd);

			dd.writeShort(methods_count);
			writeMethods(dd);

			dd.writeShort(attributes_count);
			if (attributes_count>0) {
				writeAttributes(dd,attributes_count,attributes);
			}
		} catch(IOException e) {
			System.out.println("IOException with " + fn + ": " + e.getMessage());
			return false;
		}
		return true;
	}

	// attempts to add room for more constants to the end of the
	// constant pool.  (Remember that CONSTANT_Long and CONSTANT_Double
	// require two entries.)
	// returns true if ok, false if couldn't allocate enough memory
	public boolean increaseConstantPool(short howmuch) {
		int j;
		cp_info newcp[];
		newcp = new cp_info[constant_pool_count+howmuch];
		System.arraycopy(constant_pool,0,newcp,0,constant_pool_count);
		for (j=0;j<constant_pool_count;j++) {
			constant_pool[j] = null;
		}
		constant_pool = newcp;
		Instruction.constant_pool = newcp;
		constant_pool_count += howmuch;
		return true;
	}

	// attempts to add room for more fields to the end of the
	// fields[] array;
	// returns true if ok, false if couldn't allocate enough memory
	public boolean increaseFields(short howmuch) {
		int j;
		field_info newfi[];
		newfi = new field_info[fields_count+howmuch];
		System.arraycopy(fields,0,newfi,0,fields_count);
		for (j=0;j<fields_count;j++) {
			fields[j] = null;
		}
		fields = newfi;
		fields_count += howmuch;
		return true;
	}

	// attempts to locate a Utf8 entry in the constant pool matching the
	// given string.  Returns the index in the constant pool of the entry
	// or 0 if not found
	public short locateUtf8(String s) {
		int i;
		// note that we start at 1 in the constant pool
		for (i=1;i<constant_pool_count;i++) {
			if ((constant_pool[i]).tag==cp_info.CONSTANT_Utf8) {
				CONSTANT_Utf8_info cf = (CONSTANT_Utf8_info)(constant_pool[i]);
				if (s.compareTo(cf.convert())==0) return (short)i;
			} else if ((constant_pool[i]).tag==cp_info.CONSTANT_Long || 
			           (constant_pool[i]).tag==cp_info.CONSTANT_Double) {
				// must skip an entry after a long or double constant
				i++;
			}
		}
		return (short)0;
	}

	// attempts to locate a NameAndType entry in the constant pool
	public short locateNameAndType(int name,int type) {
		int i;
		// note that we start at 1 in the constant pool
		for (i=1;i<constant_pool_count;i++) {
			if ((constant_pool[i]).tag==cp_info.CONSTANT_NameAndType) {
				CONSTANT_NameAndType_info cf = (CONSTANT_NameAndType_info)(constant_pool[i]);
				if (cf.name_index==name && cf.descriptor_index==type)
					return (short)i;
			} else if ((constant_pool[i]).tag==cp_info.CONSTANT_Long || 
			           (constant_pool[i]).tag==cp_info.CONSTANT_Double) {
				// must skip an entry after a long or double constant
				i++;
			}
		}
		return (short)0;
	}

	// attempts to locate a NameAndType entry in the constant pool
	public short locateClass(int name) {
		int i;
		// note that we start at 1 in the constant pool
		for (i=1;i<constant_pool_count;i++) {
			if ((constant_pool[i]).tag==cp_info.CONSTANT_Class) {
				CONSTANT_Class_info ci = (CONSTANT_Class_info)(constant_pool[i]);
				if (ci.name_index==name)
					return (short)i;
			} else if ((constant_pool[i]).tag==cp_info.CONSTANT_Long || 
			           (constant_pool[i]).tag==cp_info.CONSTANT_Double) {
				// must skip an entry after a long or double constant
				i++;
			}
		}
		return (short)0;
	}

	// returns a new CONSTANT_Utf8_info for the given string
	public CONSTANT_Utf8_info newUtf8(String s) {
		CONSTANT_Utf8_info cu = new CONSTANT_Utf8_info();
		cu.tag = cp_info.CONSTANT_Utf8;
		cu.bytes = CONSTANT_Utf8_info.toUtf8(s);
		//cu.length = cu.bytes.length;
		return cu;
	}

	// adds a new field reference.  Returns the index of the new field_ref
	// or 0 on error. Pass null as classname to refer to this_class
	public int addFieldRef(String classname,String name,String descriptor) {
		short descindex,nameindex,howmany,cpfree,ntindex,classindex,classnameindex;

		classindex = 0;

		if (classname == null) {
			classindex = this_class;
			classnameindex = ((CONSTANT_Class_info)(constant_pool[classindex])).name_index;
		} else {
			classnameindex = locateUtf8(classname);
			if (classnameindex!=0)
				classindex = locateClass(classnameindex);
		}

		descindex = locateUtf8(descriptor);
		nameindex = locateUtf8(name);
		if (descindex!=0 && nameindex!=0)
			ntindex = locateNameAndType(descindex,nameindex);
		else
			ntindex = 0;
		
		howmany = 1;  // need at least one for the fieldref itself
		if (descindex==0) howmany++;
		if (nameindex==0) howmany++;
		if (ntindex==0) howmany++;
		if (classindex==0) {
			if (classnameindex==0) howmany++;
			howmany++;
		}

		cpfree = constant_pool_count;  // save as new index to free slots in CP
		if (!increaseConstantPool(howmany))
			return 0;

		// add classname
		if (classnameindex==0) {
			classnameindex = cpfree;
			constant_pool[cpfree++] = newUtf8(classname);
		}

		// add class constant
		if (classindex==0) {
			CONSTANT_Class_info ci = new CONSTANT_Class_info();
			ci.tag = cp_info.CONSTANT_Class;
			ci.name_index = classnameindex;
			classindex = cpfree;
			constant_pool[cpfree++] = ci;
		}

		// add descriptor
		if (descindex==0) {
			descindex = cpfree;
			constant_pool[cpfree++] = newUtf8(descriptor);
		}

		// add actual name
		if (nameindex==0) {
			nameindex = cpfree;
			constant_pool[cpfree++] = newUtf8(name);
		}

		// add name and type
		if (ntindex==0) {
			ntindex = cpfree;
			CONSTANT_NameAndType_info cn = new CONSTANT_NameAndType_info();
			cn.tag = cp_info.CONSTANT_NameAndType;
			cn.name_index = nameindex;
			cn.descriptor_index = descindex;
			constant_pool[cpfree++] = cn;
		}

		// finally, add the fieldref
		CONSTANT_Fieldref_info cf = new CONSTANT_Fieldref_info();
		cf.tag = cp_info.CONSTANT_Fieldref;
		cf.class_index = classindex;
		cf.name_and_type_index = ntindex;
		constant_pool[cpfree] = cf;

		// finished 
		return cpfree;
	}


	// adds a new method reference.  Returns the index of the new method_ref
	// or 0 on error. Pass null as classname to refer to this_class
	public int addMethodRef(String classname,String name,String descriptor) {
		short descindex,nameindex,howmany,cpfree,ntindex,classindex,classnameindex;

		classindex = 0;

		if (classname == null) {
			classindex = this_class;
			classnameindex = ((CONSTANT_Class_info)(constant_pool[classindex])).name_index;
		} else {
			classnameindex = locateUtf8(classname);
			if (classnameindex!=0)
				classindex = locateClass(classnameindex);
		}

		descindex = locateUtf8(descriptor);
		nameindex = locateUtf8(name);
		if (descindex!=0 && nameindex!=0)
			ntindex = locateNameAndType(descindex,nameindex);
		else
			ntindex = 0;
		
		howmany = 1;  // need at least one for the fieldref itself
		if (descindex==0) howmany++;
		if (nameindex==0) howmany++;
		if (ntindex==0) howmany++;
		if (classindex==0) {
			if (classnameindex==0) howmany++;
			howmany++;
		}

		cpfree = constant_pool_count;  // save as new index to free slots in CP
		if (!increaseConstantPool(howmany))
			return 0;

		// add classname
		if (classnameindex==0) {
			classnameindex = cpfree;
			constant_pool[cpfree++] = newUtf8(classname);
		}

		// add class constant
		if (classindex==0) {
			CONSTANT_Class_info ci = new CONSTANT_Class_info();
			ci.tag = cp_info.CONSTANT_Class;
			ci.name_index = classnameindex;
			classindex = cpfree;
			constant_pool[cpfree++] = ci;
		}

		// add descriptor
		if (descindex==0) {
			descindex = cpfree;
			constant_pool[cpfree++] = newUtf8(descriptor);
		}

		// add actual name
		if (nameindex==0) {
			nameindex = cpfree;
			constant_pool[cpfree++] = newUtf8(name);
		}

		// add name and type
		if (ntindex==0) {
			ntindex = cpfree;
			CONSTANT_NameAndType_info cn = new CONSTANT_NameAndType_info();
			cn.tag = cp_info.CONSTANT_NameAndType;
			cn.name_index = nameindex;
			cn.descriptor_index = descindex;
			constant_pool[cpfree++] = cn;
		}

		// finally, add the methodref
		CONSTANT_Methodref_info cf = new CONSTANT_Methodref_info();
		cf.tag = cp_info.CONSTANT_Methodref;
		cf.class_index = classindex;
		cf.name_and_type_index = ntindex;
		constant_pool[cpfree] = cf;

		// finished 
		return cpfree;
	}

	// inserts the given sequence of instructions into the given code after
	// the given instruction.  Use null for after to insert prior to any code.
	public boolean insertCode(method_info m,Instruction insrt,Instruction after) {
		Instruction i,afteri,lastinsrt;
		Code_attribute ca;
		int index;
		
		ca = locate_code_attribute(m);
		if (ca==null) {
			System.out.println("Warning: No code attribute found in method");
			return false;
		}
		// first determine how much space the new chunk of code requires
		// note that because of some instructions being dword aligned, 
		// we have to renumber the instructions to determine how much
		// extra space we'll need.
		index = 0;
		afteri = m.instructions;
		if (after!=null) {
			while (afteri!=after) {
				index = afteri.nextOffset(index);
				afteri = afteri.next;
			} 
			index = afteri.nextOffset(index);
			afteri = afteri.next;
		}
		i = insrt;
		lastinsrt = insrt;
		while (i!=null) {
			index = i.nextOffset(index);
			lastinsrt = i;
			i = i.next;
		}
		while (afteri!=null) {
			index = afteri.nextOffset(index);
			afteri = afteri.next;
		}
		
		// 
		// see if there's enough room for that much code
		if (index > 32767) {
			if (index > Integer.MAX_VALUE) {
				System.out.println("Warning: can't insert an extra " + 
								   (index-ca.code_length)  + 
				                   " bytes of code into " + ca.code_length + 
				                   " bytes of code");
				return false;
			}
			System.out.println("Warning: inserting " + (index-ca.code_length) + 
			                   " bytes of code may cause offsets to be unreachable...");
		}
		
		// locate end of code sequence to insert
		if (after==null) {
			i.next = m.instructions;
			m.instructions = insrt;
		} else {
			lastinsrt.next = after.next;
			after.next = insrt;
		}

		// finally, update size of the code attribute
		ca.attribute_length += index-ca.code_length;

		// that should be it; recompile will take care of the rest
		// note that we have not altered the code_length, nor the max_stack,
		// nor any of the debugging attributes.
		return true;
	}

	public boolean addField(short access,String name,String descriptor) {
		field_info fi = new field_info();
		short howmany,cpfree,fieldindex;

		fieldindex = fields_count;
		if (!increaseFields((short)1))
			return false;

		fi.descriptor_index = locateUtf8(descriptor);
		if (fi.descriptor_index!=0) howmany = 1; // 1 for field name
		else howmany = 2;              // 1 more entry for descriptor

		cpfree = constant_pool_count;  // save as new index to free slots in CP
		if (!increaseConstantPool(howmany))
			return false;
		// first add descriptor if necessary
		if (fi.descriptor_index==0) {
			fi.descriptor_index = cpfree;
			constant_pool[cpfree++] = newUtf8(descriptor);
		}

		// add actual name
		fi.name_index = cpfree;
		constant_pool[cpfree++] = newUtf8(name);

		fi.access_flags = access;
		fi.attributes_count = 0;
		fi.attributes = null;

		fields[fieldindex] = fi;
		// finished 
		return true;
	}

	// converts byte code into a list of instructions 
	public Instruction decompile(method_info m) {
		// first task, look through attributes for a code attribute
		int j;
		Code_attribute ca;
		ByteCode bc;
		Instruction inst,head,tail;
		exception_table_entry e;

		head = null;
		tail = null;
		bc = new ByteCode();

		ca = locate_code_attribute(m);
		if (ca==null) return null;

		//System.out.println(Integer.toString(ca.code_length) + " bytes of code follows...");
		//showCode(ca);
		j = 0;
		while(j<ca.code_length) {
			//System.out.print(Integer.toString(j) + ": ");
			inst = bc.disassemble_bytecode(ca.code,j);
			if (inst instanceof Instruction_Unknown) {
				System.out.println("Unknown instruction in \"" + m.toName(constant_pool) +
								   "\" at offset " + j);
				System.out.println(" bytecode = " + (((int)(inst.code))&0xff));
			}
			j = inst.nextOffset(j);
			if (head==null) head = inst;
			else tail.next = inst;
			tail = inst;
			//System.out.println(bc.i.toString());
		}

		// bytecode converted into instructions, now build pointers
		bc.build(head);

		// also change exception table to use pointers instead of absolute addresses
		for (j=0;j<ca.exception_table_length;j++) {
			e = ca.exception_table[j];
			e.start_inst = bc.locateInst(e.start_pc);
			if (e.end_pc == ca.code_length)
				e.end_inst = null;
			else
				e.end_inst = bc.locateInst(e.end_pc);
			e.handler_inst = bc.locateInst(e.handler_pc);
			if (e.handler_inst!=null)
				e.handler_inst.labelled = true;
		}

		return head;
	}

	// decompiles every method
	public void parse() {
		method_info mi;
		int i;

		for (i=0;i<methods_count;i++) {
			mi = methods[i];
			mi.instructions = decompile(mi);
		}
	}

	// recomputes the label of each instruction.  Returns the total size.
	public int relabel(Instruction i) {
		int index = 0;
		while (i!=null) {
			i.label = index;
			index = i.nextOffset(index);
			i = i.next;
		}
		return index;
	}

	// converts list of instructions into byte code 
	public boolean recompile(method_info m) {
		// first task, look through attributes for a code attribute
		int codesize,j;
		Code_attribute ca;
		byte bc[];
		Instruction i;
		exception_table_entry e;

		// locate code attribute
		ca = locate_code_attribute(m);
		if (ca==null) return false;

		// relabel instructions 
		codesize = relabel(m.instructions);

		// construct a new array for the byte-code
		bc = new byte[codesize];
		if (bc==null) {
			System.out.println("Warning: can't allocate memory for recompile");
			return false;
		}
		
		// then recompile the instructions into byte-code
		i = m.instructions;
		codesize = 0;
		while (i!=null) {
			codesize = i.compile(bc,codesize);
			i = i.next;
		}
		if (codesize != bc.length)
			System.out.println("Warning: code size doesn't match array length!");
		ca.code_length = codesize;
		ca.code = bc;

		// also recompile exception table
		for (j=0;j<ca.exception_table_length;j++) {
			e = ca.exception_table[j];
			e.start_pc = (short)(e.start_inst.label);
			if (e.end_inst!=null)
				e.end_pc = (short)(e.end_inst.label);
			else
				e.end_pc = (short)(ca.code_length);
			e.handler_pc = (short)(e.handler_inst.label);
		}

		return true;
	}

	public void recompileAll() {
		int i;
		method_info mi;

		for (i=0;i<methods_count;i++) {
			mi = methods[i];
			if (!recompile(mi)) {
				System.out.println("Recompile of " + mi.toName(constant_pool) + " failed!");
			}
		}

	}

	// returns the return type of the given method descriptor
	public String parseMethodDesc_return(String s) {
		int j;
		j = s.lastIndexOf(')');
		if (j>=0) {
			return parseDesc(s.substring(j+1),",");
		}
		return parseDesc(s,",");
	}

	// returns a list of parameters from the given method descriptor
	public String parseMethodDesc_params(String s) {
		int i,j;
		i = s.indexOf('(');
		if (i>=0) {
			j = s.indexOf(')',i+1);
			if (j>=0) {
				return parseDesc(s.substring(i+1,j),",");
			}
		}
		return "<parse error>";
	}

	// Returns a string of types separated by sep given the descriptor string
	public String parseDesc(String desc,String sep) {
		String params = "",param;
		char c;
		int i,len,arraylevel=0;
		boolean didone = false;

		len = desc.length();
		for (i=0;i<len;i++) {
			c = desc.charAt(i);
			if (c==DESC_BYTE.charAt(0)) {
				param = "byte";
			} else if (c==DESC_CHAR.charAt(0)) {
				param = "char";
			} else if (c==DESC_DOUBLE.charAt(0)) {
				param = "double";
			} else if (c==DESC_FLOAT.charAt(0)) {
				param = "float";
			} else if (c==DESC_INT.charAt(0)) {
				param = "int";
			} else if (c==DESC_LONG.charAt(0)) {
				param = "long";
			} else if (c==DESC_SHORT.charAt(0)) {
				param = "short";
			} else if (c==DESC_BOOLEAN.charAt(0)) {
				param = "boolean";
			} else if (c==DESC_VOID.charAt(0)) {
				param = "void";
			} else if (c==DESC_ARRAY.charAt(0)) {
				arraylevel++;
				continue;
			} else if (c==DESC_OBJECT.charAt(0)) {
				int j;
				j = desc.indexOf(';',i+1);
				if (j<0) {
					System.out.println("Warning: Parse error -- can't find a ; in " + 
					                   desc.substring(i+1));
					param = "<error>";
				} else {
					if (j-i>10 && desc.substring(i+1,i+11).compareTo("java/lang/")==0)
						i = i+10;
					param = desc.substring(i+1,j);
					// replace '/'s with '.'s
					param = param.replace('/','.');
					i = j;
				}
			} else	{
				param = "???";
			}
			if (didone) params = params + sep;
			params = params + param;
			while (arraylevel>0) {
				params = params + "[]";
				arraylevel--;
			}
			didone = true;
		}
		return params;
	}

	/* // for building the profiling code
	public boolean addProfileCode() {
		Instruction head,tail;
		int i,hashref,hashput,hashget;
		method_info m;
		Code_attribute ca;

		// add a hashtable as a new field
		addField(ACC_PUBLIC,HASHTABLEVAR,HASHTABLEDESC);

		// add the corresponding field reference, and store it for later
		hashref = addFieldRef(null,HASHTABLEVAR,HASHTABLEDESC);

		// we'll also want to be able to call put and get 
		hashput = addMethodRef(HASHTABLE,HASHTABLEPUT,
		                       "("+OBJECTDESC+";"+OBJECTDESC+";)"+OBJECTDESC);

		// we'll also want to be able to call put and get 
		hashget = addMethodRef(HASHTABLE,HASHTABLEGET,
		                       "("+OBJECTDESC+";)"+OBJECTDESC);

		// locate shortest method
		bestm = methods[0];
		bestca = locate_code_attribute(bestm);
		for (i=1;i<methods_count;i++) {
			ca = locate_code_attribute(methods[i]);
			if (ca!=null && ca.code_length<bestca.code_length) {
				bestca = ca;
				bestm = methods[i];
			}
		}
		System.out.println("Testing insertion of code into " + bestm.toName(constant_pool));
		System.out.println("Here's what the code looked like before: ");
		ByteCode.showCode(bestm.instructions);
		// make a silly sequence of code
		donothing = new Instruction_Bipush((byte)7);
		donothing.next = new Instruction_Pop();
		
		if (!insertCode(bestm,donothing,bestm.instructions.next.next))
			System.out.println("Can't insert code!");
		System.out.println("Here's what the code looks like after: ");
		ByteCode.showCode(bestm.instructions);



		// start code looks something like:
		// getfield this.
		cpfree = constant_pool_count;  // save as new index to free slots in CP
		if (!increaseConstantPool(howmany))
			return false;
		// first add descriptor if necessary
		if (fi.descriptor_index==0) {
			fi.descriptor_index = cpfree;
			constant_pool[cpfree++] = newUtf8(descriptor);
		}


		head = new Instruction_Getfield();
		((Instruction_Getfield)head).arg_i = 
	}*/

	public void showMethod(String s) {
		method_info m;
		CONSTANT_Utf8_info cm;
		int i;

		for (i=0;i<methods_count;i++) {
			m = methods[i];
			cm = (CONSTANT_Utf8_info)(constant_pool[m.name_index]);
			if (s.compareTo(cm.convert())==0) {
				System.out.println("Method: " + s);
				System.out.println("Signature: " + ((CONSTANT_Utf8_info)
				                   (constant_pool[m.descriptor_index])).convert());
				ByteCode.showCode(m.instructions);
				return;
			}
		}
		System.out.println("Warning: method " + s + " not found!");
	}

	public void showCFG(String s) {
		method_info m;
		CONSTANT_Utf8_info cm;
		int i;

		for (i=0;i<methods_count;i++) {
			m = methods[i];
			cm = (CONSTANT_Utf8_info)(constant_pool[m.name_index]);
			if (s.compareTo(cm.convert())==0) {
				CFG cfg = new CFG(m);
				cfg.buildCFG();
				cfg.showCFG(s);
				return;
			}
		}
		System.out.println("Warning: method " + s + " not found!");
	}

	// for testing features
	public void testit() {
		// try and insert a simple, do-nothing chunk of code
		int i,hashref;
		Instruction donothing;
		method_info bestm;
		Code_attribute ca,bestca;

		// try adding a hashtable as a new field
		addField(ACC_PUBLIC,"hey_im_a_hashtable",DESC_OBJECT + "java/util/Hashtable");

		// and also adding a field reference for that field
		hashref = addFieldRef(null,"hey_im_a_hashtable",DESC_OBJECT + "java/util/Hashtable");

		// locate shortest method
		bestm = methods[0];
		bestca = locate_code_attribute(bestm);
		for (i=1;i<methods_count;i++) {
			ca = locate_code_attribute(methods[i]);
			if (ca!=null && ca.code_length<bestca.code_length) {
				bestca = ca;
				bestm = methods[i];
			}
		}
		System.out.println("Testing insertion of code into " + bestm.toName(constant_pool));
		System.out.println("Here's what the code looked like before: ");
		ByteCode.showCode(bestm.instructions);
		// make a silly sequence of code
		donothing = new Instruction_Bipush((byte)7);
		donothing.next = new Instruction_Pop();
		
		if (!insertCode(bestm,donothing,bestm.instructions.next.next))
			System.out.println("Can't insert code!");
		System.out.println("Here's what the code looks like after: ");
		ByteCode.showCode(bestm.instructions);
	}

	public void listMethods() {
		method_info m;
		String s;
		CONSTANT_Utf8_info cm,dm;
		int i;

		for (i=0;i<methods_count;i++) {
			m = methods[i];
			cm = (CONSTANT_Utf8_info)(constant_pool[m.name_index]);
			dm = (CONSTANT_Utf8_info)(constant_pool[m.descriptor_index]);
			System.out.print(access_string(m.access_flags," ").toLowerCase() + " ");
			s = dm.convert();
			System.out.print(parseMethodDesc_return(s) + " ");
			System.out.print(cm.convert() + "(");
			System.out.println(parseMethodDesc_params(s) + ")");
		}
	}

	public void listConstantPool() {
		cp_info c;
		int i;

		// note that we start at 1 in the constant pool
		for (i=1;i<constant_pool_count;i++) {
			c = constant_pool[i];
			System.out.println("[" + i + "] " + c.typeName() + "=" + c.toString(constant_pool));
			if ((constant_pool[i]).tag==cp_info.CONSTANT_Long || 
			    (constant_pool[i]).tag==cp_info.CONSTANT_Double) {
				// must skip an entry after a long or double constant
				i++;
			}
		}
	}

	public void listFields() {
		field_info fi;
		ConstantValue_attribute cva;
		CONSTANT_Utf8_info cm,dm;
		int i,j;
		String s;

		for (i=0;i<fields_count;i++) {
			fi = fields[i];
			cm = (CONSTANT_Utf8_info)(constant_pool[fi.name_index]);
			dm = (CONSTANT_Utf8_info)(constant_pool[fi.descriptor_index]);
			s = access_string(fi.access_flags," ").toLowerCase();
			if (s.compareTo("")!=0) s = s + " ";
			System.out.print(s + parseDesc(dm.convert(),"") + " " + cm.convert());
			s = ";";
			// see if has a constant value attribute
			for (j=0;j<fi.attributes_count;j++) {
				cm = (CONSTANT_Utf8_info)(constant_pool[fi.attributes[j].attribute_name]);
				if (cm.convert().compareTo(attribute_info.ConstantValue)==0) {
					cva = (ConstantValue_attribute)(fi.attributes[j]);
					//dm = (CONSTANT_Utf8_info)(constant_pool[cva.constantvalue_index]);
					System.out.print(" = " + 
					       constant_pool[cva.constantvalue_index].toString(constant_pool));
					break;
				}
			}
			System.out.println(";");
		}
	}
}
