/*
 Java Class Inspector
   Copyright (C) 1996 Clark VERBRUGGE.     Contact: clump@cs.mcgill.ca
   Copyright (C) 1996 McGill University. 
   Copyright (C) 1996 McCAT System Group.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 1, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 675
   Mass Ave, Cambridge, MA 02139, USA. */

import java.io.*;

public abstract class Instruction {

	public static final String argsep = " ";  // String used to separate arguments
	public static final String LOCALPREFIX = "temp_"; // for naming local variables

	public static cp_info constant_pool[];
	
	public static int w; // set by the wide instr. and used by other instrs

	byte code;           // actual byte code
	int label;           // offset from start of code
	// int size;            // number of bytes required to express this instruction
	String name;         // name of code
	Instruction next;    // for chaining
	boolean labelled;    // whether instruction is a target of a branch
	boolean branches;    // whether instruction can branch or not
	boolean calls;       // whether instruction makes a function call
	boolean returns;     // whether instruction is a return

	Instruction(byte c) { 
		code = c; 
		next = null; 
		branches = false; 
		calls = false; 
		returns = false; 
	}

	// Assuming the code has been extracted already, and index is on the next byte
	// this function must interpret whatever arguments the instruction requires
	// and return the index of the next available byte
	abstract int parse(byte bc[],int index);

	// Undoes the work of parse; this converts this instruction into 
	// a sequence of byte codes and returns the offset of the next 
	// free spot
	abstract int compile(byte bc[],int index);

	// Changes any offset references to pointers
	// default behaviour, since most instructions are not branches,
	// is just to ignore the issue
	void offsetToPointer(ByteCode bc) { }

	// Returns the next available offset if this instruction starts on 
	// the indicated byte
	int nextOffset(int curr) { return curr+1; }

	// Returns a list of all instructions this might branch to
	Instruction[] branchpoints(Instruction next) { return null; }

	// for storing in a hashtable
	public int hashCode() {
		return (new Integer(label)).hashCode();
	}

	public boolean equals(Instruction i) {
		if (label == i.label) return true;
		return false;
	}

	// converts the next 2 bytes into a short value
	public static short getShort(byte bc[],int index) {
		short s,bh,bl;
		bh = (short)(bc[index]); bl = (short)(bc[index+1]);
		s = (short)(((bh<<8)&0xff00) | (bl&0xff));
		//s = (short)((int)(bc[index])<<8 + bc[index+1]);	
		return s;
	}
	public static int getInt(byte bc[],int index) {
		int i,bhh,bhl,blh,bll;
		bhh = (((int)(bc[index]))<<24)&0xff000000;
		bhl = (((int)(bc[index+1]))<<16)&0xff0000;
		blh = (((int)(bc[index+2]))<<8)&0xff00;
		bll = ((int)(bc[index+3]))&0xff;
		i = bhh | bhl | blh | bll;
		return i;
	}

	// converts the given short s into two unsigned bytes and stores them
	// (mfb first) in bc
	public static int shortToBytes(short s,byte bc[],int index) {
		bc[index++] = (byte)((s>>8)&0xff);
		bc[index++] = (byte)(s&0xff);
		return index;
	}

	// converts the given short s into four unsigned bytes and stores them
	// (mfb first) in bc
	public static int intToBytes(int s,byte bc[],int index) {
		bc[index++] = (byte)((s>>24)&0xff);
		bc[index++] = (byte)((s>>16)&0xff);
		bc[index++] = (byte)((s>>8)&0xff);
		bc[index++] = (byte)(s&0xff);
		return index;
	}

	public String toString() {
		int i = ((int)code)&0xff;
		if (name==null) name = "null???=" + Integer.toString(i);
		return name;
	}
}
class Instruction_noargs extends Instruction {
	Instruction_noargs(byte c) { super(c); }
	int parse(byte bc[],int index) { return index; }
	int compile(byte bc[],int index) { bc[index++] = code; return index; }
}
class Instruction_byte extends Instruction {
	byte arg_b;
	Instruction_byte(byte c) { super(c); }
	public String toString() {
		return super.toString() + argsep + arg_b;
	}
	int nextOffset(int curr) { return curr+2; }
	int parse(byte bc[],int index) { arg_b = bc[index]; return index+1; }
	int compile(byte bc[],int index) { bc[index++] = code; bc[index++] = arg_b; return index; }
}
class Instruction_byteindex extends Instruction {
	byte arg_b;
	Instruction_byteindex(byte c) { super(c); }
	public String toString() {
		int i = ((int)arg_b)&0xff;
		return super.toString() + argsep + "[" + constant_pool[i].toString(constant_pool) + "]";
	}
	int nextOffset(int curr) { return curr+2; }
	int parse(byte bc[],int index) { 
		arg_b = bc[index]; 
		return index+1; 
	}
	int compile(byte bc[],int index) { bc[index++] = code; bc[index++] = arg_b; return index; }
}
class Instruction_bytevar extends Instruction {
	byte arg_b;
	Instruction_bytevar(byte c) { super(c); }
	public String toString() {
		return super.toString() + argsep + LOCALPREFIX + arg_b;
	}
	int nextOffset(int curr) { return curr+2; }
	int parse(byte bc[],int index) { 
		arg_b = bc[index]; 
		return index+1; 
	}
	int compile(byte bc[],int index) { bc[index++] = code; bc[index++] = arg_b; return index; }
}
class Instruction_int extends Instruction {
	int arg_i;
	Instruction_int(byte c) { super(c); }
	public String toString() {
		return super.toString() + argsep + arg_i;
	}
	int nextOffset(int curr) { return curr+3; }
	int parse(byte bc[],int index) { 
		arg_i = getShort(bc,index);
		return index+2; 
	}
	int compile(byte bc[],int index) { 
		bc[index++] = code; 
		shortToBytes((short)arg_i,bc,index); 
		return index+2; 
	}
}
class Instruction_intindex extends Instruction {
	int arg_i;
	Instruction_intindex(byte c) { super(c); }
	public String toString() {
		return super.toString() + argsep + "[" + 
		       constant_pool[arg_i].toString(constant_pool) + "]";
	}
	int nextOffset(int curr) { return curr+3; }
	int parse(byte bc[],int index) { 
		arg_i = getShort(bc,index);
		return index+2; 
	}
	int compile(byte bc[],int index) { 
		bc[index++] = code; 
		shortToBytes((short)arg_i,bc,index); 
		return index+2; 
	}
}
class Instruction_intvar extends Instruction {
	int arg_i;
	Instruction_intvar(byte c) { super(c); }
	public String toString() {
		return super.toString() + argsep + LOCALPREFIX + arg_i;
	}
	int nextOffset(int curr) { return curr+3; }
	int parse(byte bc[],int index) { 
		arg_i = getShort(bc,index);
		return index+2; 
	}
	int compile(byte bc[],int index) { 
		bc[index++] = code; 
		shortToBytes((short)arg_i,bc,index); 
		return index+2; 
	}
}
class Instruction_intbranch extends Instruction {
	int arg_i;
	Instruction target;         // pointer to target instruction
	Instruction_intbranch(byte c) { super(c); branches = true; }
	public String toString() {
		return super.toString() + argsep + "[label_" + 
		  Integer.toString(target.label) + "]";
	}
	int nextOffset(int curr) { return curr+3; }
	int parse(byte bc[],int index) { 
		arg_i = getShort(bc,index);
		return index+2; 
	}
	int compile(byte bc[],int index) { 
		bc[index++] = code; 
		if (target!=null)
			shortToBytes((short)(target.label-label),bc,index);
		else
			shortToBytes((short)(arg_i),bc,index); 
		return index+2; 
	}
	void offsetToPointer(ByteCode bc) { 
		target = bc.locateInst(arg_i+label);
		if (target==null) {
			System.out.println("Warning: can't locate target of instruction");
			System.out.println(" which should be at byte address " + (label+arg_i));
		} else
		  target.labelled = true;
	}
	// returns the array of instructions which might be the target of a
	// branch with this instruction, assuming the next instruction is next
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[2];
		i[0] = target; i[1] = next;
		return i;
	}
}
class Instruction_longbranch extends Instruction {
	int arg_i;
	Instruction target;
	Instruction_longbranch(byte c) { super(c); branches = true; }
	public String toString() {
		return super.toString() + argsep + "[label_" + Integer.toString(target.label) + "]";
	}
	int nextOffset(int curr) { return curr+5; }
	int parse(byte bc[],int index) { 
		arg_i = getInt(bc,index);
		return index+4; 
	}
	int compile(byte bc[],int index) { 
		bc[index++] = code; 
		if (target!=null)
			shortToBytes((short)(target.label-label),bc,index);
		else
			shortToBytes((short)(arg_i),bc,index); 
		return index+4; 
	}
	void offsetToPointer(ByteCode bc) { 
		target = bc.locateInst(arg_i+label);
		if (target==null) {
			System.out.println("Warning: can't locate target of instruction");
			System.out.println(" which should be at byte address " + (label+arg_i));
		} else
		  target.labelled = true;
	}
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[2];
		i[0] = target; i[1] = next;
		return i;
	}
}

class Instruction_Unknown extends Instruction_noargs {
	Instruction_Unknown(byte c) { 
		super(c); 
		name = "unknown instruction (" + (((int)c)&0xff) + ")"; 
	}
}
class Instruction_Bipush extends Instruction_byte {
	Instruction_Bipush() { super((byte)ByteCode.BIPUSH); name = "bipush"; }
	Instruction_Bipush(byte b) { this(); arg_b = b; }
}
class Instruction_Sipush extends Instruction_int {
	Instruction_Sipush() { super((byte)ByteCode.SIPUSH); name = "sipush"; }
	Instruction_Sipush(int i) { this(); arg_i = i; }
}
class Instruction_Ldc1 extends Instruction_byteindex {
	Instruction_Ldc1() { super((byte)ByteCode.LDC1); name = "ldc1"; }
}
class Instruction_Ldc2 extends Instruction_intindex {
	Instruction_Ldc2() { super((byte)ByteCode.LDC2); name = "ldc2"; }
}
class Instruction_Ldc2w extends Instruction_intindex {
	Instruction_Ldc2w() { super((byte)ByteCode.LDC2W); name = "ldc2w"; }
}
class Instruction_Aconst_null extends Instruction_noargs {
	Instruction_Aconst_null() { super((byte)ByteCode.ACONST_NULL); name = "aconst_null"; }
}
class Instruction_Iconst_m1 extends Instruction_noargs {
	Instruction_Iconst_m1() { super((byte)ByteCode.ICONST_M1); name = "iconst_m1"; }
}
class Instruction_Iconst_0 extends Instruction_noargs {
	Instruction_Iconst_0() { super((byte)ByteCode.ICONST_0); name = "iconst_0"; }
}
class Instruction_Iconst_1 extends Instruction_noargs {
	Instruction_Iconst_1() { super((byte)ByteCode.ICONST_1); name = "iconst_1"; }
}
class Instruction_Iconst_2 extends Instruction_noargs {
	Instruction_Iconst_2() { super((byte)ByteCode.ICONST_2); name = "iconst_2"; }
}
class Instruction_Iconst_3 extends Instruction_noargs {
	Instruction_Iconst_3() { super((byte)ByteCode.ICONST_3); name = "iconst_3"; }
}
class Instruction_Iconst_4 extends Instruction_noargs {
	Instruction_Iconst_4() { super((byte)ByteCode.ICONST_4); name = "iconst_4"; }
}
class Instruction_Iconst_5 extends Instruction_noargs {
	Instruction_Iconst_5() { super((byte)ByteCode.ICONST_5); name = "iconst_5"; }
}
class Instruction_Lconst_0 extends Instruction_noargs {
	Instruction_Lconst_0() { super((byte)ByteCode.LCONST_0); name = "lconst_0"; }
}
class Instruction_Lconst_1 extends Instruction_noargs {
	Instruction_Lconst_1() { super((byte)ByteCode.LCONST_1); name = "lconst_1"; }
}
class Instruction_Fconst_0 extends Instruction_noargs {
	Instruction_Fconst_0() { super((byte)ByteCode.FCONST_0); name = "fconst_0"; }
}
class Instruction_Fconst_1 extends Instruction_noargs {
	Instruction_Fconst_1() { super((byte)ByteCode.FCONST_1); name = "fconst_1"; }
}
class Instruction_Fconst_2 extends Instruction_noargs {
	Instruction_Fconst_2() { super((byte)ByteCode.FCONST_2); name = "fconst_2"; }
}
class Instruction_Dconst_0 extends Instruction_noargs {
	Instruction_Dconst_0() { super((byte)ByteCode.DCONST_0); name = "dconst_0"; }
}
class Instruction_Dconst_1 extends Instruction_noargs {
	Instruction_Dconst_1() { super((byte)ByteCode.DCONST_1); name = "dconst_1"; }
}
class Instruction_Iload extends Instruction_bytevar {
	Instruction_Iload() { super((byte)ByteCode.ILOAD); name = "iload"; }
}
class Instruction_Iload_0 extends Instruction_noargs {
	Instruction_Iload_0() { super((byte)ByteCode.ILOAD_0); name = "iload_0"; }
}
class Instruction_Iload_1 extends Instruction_noargs {
	Instruction_Iload_1() { super((byte)ByteCode.ILOAD_1); name = "iload_1"; }
}
class Instruction_Iload_2 extends Instruction_noargs {
	Instruction_Iload_2() { super((byte)ByteCode.ILOAD_2); name = "iload_2"; }
}
class Instruction_Iload_3 extends Instruction_noargs {
	Instruction_Iload_3() { super((byte)ByteCode.ILOAD_3); name = "iload_3"; }
}
class Instruction_Lload extends Instruction_bytevar {
	Instruction_Lload() { super((byte)ByteCode.LLOAD); name = "lload"; }
}
class Instruction_Lload_0 extends Instruction_noargs {
	Instruction_Lload_0() { super((byte)ByteCode.LLOAD_0); name = "lload_0"; }
}
class Instruction_Lload_1 extends Instruction_noargs {
	Instruction_Lload_1() { super((byte)ByteCode.LLOAD_1); name = "lload_1"; }
}
class Instruction_Lload_2 extends Instruction_noargs {
	Instruction_Lload_2() { super((byte)ByteCode.LLOAD_2); name = "lload_2"; }
}
class Instruction_Lload_3 extends Instruction_noargs {
	Instruction_Lload_3() { super((byte)ByteCode.LLOAD_3); name = "lload_3"; }
}
class Instruction_Fload extends Instruction_bytevar {
	Instruction_Fload() { super((byte)ByteCode.FLOAD); name = "fload"; }
}
class Instruction_Fload_0 extends Instruction_noargs {
	Instruction_Fload_0() { super((byte)ByteCode.FLOAD_0); name = "fload_0"; }
}
class Instruction_Fload_1 extends Instruction_noargs {
	Instruction_Fload_1() { super((byte)ByteCode.FLOAD_1); name = "fload_1"; }
}
class Instruction_Fload_2 extends Instruction_noargs {
	Instruction_Fload_2() { super((byte)ByteCode.FLOAD_2); name = "fload_2"; }
}
class Instruction_Fload_3 extends Instruction_noargs {
	Instruction_Fload_3() { super((byte)ByteCode.FLOAD_3); name = "fload_3"; }
}
class Instruction_Dload extends Instruction_bytevar {
	Instruction_Dload() { super((byte)ByteCode.DLOAD); name = "dload"; }
}
class Instruction_Dload_0 extends Instruction_noargs {
	Instruction_Dload_0() { super((byte)ByteCode.DLOAD_0); name = "dload_0"; }
}
class Instruction_Dload_1 extends Instruction_noargs {
	Instruction_Dload_1() { super((byte)ByteCode.DLOAD_1); name = "dload_1"; }
}
class Instruction_Dload_2 extends Instruction_noargs {
	Instruction_Dload_2() { super((byte)ByteCode.DLOAD_2); name = "dload_2"; }
}
class Instruction_Dload_3 extends Instruction_noargs {
	Instruction_Dload_3() { super((byte)ByteCode.DLOAD_3); name = "dload_3"; }
}
class Instruction_Aload extends Instruction_bytevar {
	Instruction_Aload() { super((byte)ByteCode.ALOAD); name = "aload"; }
}
class Instruction_Aload_0 extends Instruction_noargs {
	Instruction_Aload_0() { super((byte)ByteCode.ALOAD_0); name = "aload_0"; }
}
class Instruction_Aload_1 extends Instruction_noargs {
	Instruction_Aload_1() { super((byte)ByteCode.ALOAD_1); name = "aload_1"; }
}
class Instruction_Aload_2 extends Instruction_noargs {
	Instruction_Aload_2() { super((byte)ByteCode.ALOAD_2); name = "aload_2"; }
}
class Instruction_Aload_3 extends Instruction_noargs {
	Instruction_Aload_3() { super((byte)ByteCode.ALOAD_3); name = "aload_3"; }
}
class Instruction_Istore extends Instruction_bytevar {
	Instruction_Istore() { super((byte)ByteCode.ISTORE); name = "istore"; }
}
class Instruction_Istore_0 extends Instruction_noargs {
	Instruction_Istore_0() { super((byte)ByteCode.ISTORE_0); name = "istore_0"; }
}
class Instruction_Istore_1 extends Instruction_noargs {
	Instruction_Istore_1() { super((byte)ByteCode.ISTORE_1); name = "istore_1"; }
}
class Instruction_Istore_2 extends Instruction_noargs {
	Instruction_Istore_2() { super((byte)ByteCode.ISTORE_2); name = "istore_2"; }
}
class Instruction_Istore_3 extends Instruction_noargs {
	Instruction_Istore_3() { super((byte)ByteCode.ISTORE_3); name = "istore_3"; }
}
class Instruction_Lstore extends Instruction_bytevar {
	Instruction_Lstore() { super((byte)ByteCode.LSTORE); name = "lstore"; }
}
class Instruction_Lstore_0 extends Instruction_noargs {
	Instruction_Lstore_0() { super((byte)ByteCode.LSTORE_0); name = "lstore_0"; }
}
class Instruction_Lstore_1 extends Instruction_noargs {
	Instruction_Lstore_1() { super((byte)ByteCode.LSTORE_1); name = "lstore_1"; }
}
class Instruction_Lstore_2 extends Instruction_noargs {
	Instruction_Lstore_2() { super((byte)ByteCode.LSTORE_2); name = "lstore_2"; }
}
class Instruction_Lstore_3 extends Instruction_noargs {
	Instruction_Lstore_3() { super((byte)ByteCode.LSTORE_3); name = "lstore_3"; }
}
class Instruction_Fstore extends Instruction_bytevar {
	Instruction_Fstore() { super((byte)ByteCode.FSTORE); name = "fstore"; }
}
class Instruction_Fstore_0 extends Instruction_noargs {
	Instruction_Fstore_0() { super((byte)ByteCode.FSTORE_0); name = "fstore_0"; }
}
class Instruction_Fstore_1 extends Instruction_noargs {
	Instruction_Fstore_1() { super((byte)ByteCode.FSTORE_1); name = "fstore_1"; }
}
class Instruction_Fstore_2 extends Instruction_noargs {
	Instruction_Fstore_2() { super((byte)ByteCode.FSTORE_2); name = "fstore_2"; }
}
class Instruction_Fstore_3 extends Instruction_noargs {
	Instruction_Fstore_3() { super((byte)ByteCode.FSTORE_3); name = "fstore_3"; }
}
class Instruction_Dstore extends Instruction_bytevar {
	Instruction_Dstore() { super((byte)ByteCode.DSTORE); name = "dstore"; }
}
class Instruction_Dstore_0 extends Instruction_noargs {
	Instruction_Dstore_0() { super((byte)ByteCode.DSTORE_0); name = "dstore_0"; }
}
class Instruction_Dstore_1 extends Instruction_noargs {
	Instruction_Dstore_1() { super((byte)ByteCode.DSTORE_1); name = "dstore_1"; }
}
class Instruction_Dstore_2 extends Instruction_noargs {
	Instruction_Dstore_2() { super((byte)ByteCode.DSTORE_2); name = "dstore_2"; }
}
class Instruction_Dstore_3 extends Instruction_noargs {
	Instruction_Dstore_3() { super((byte)ByteCode.DSTORE_3); name = "dstore_3"; }
}
class Instruction_Astore extends Instruction_bytevar {
	Instruction_Astore() { super((byte)ByteCode.ASTORE); name = "astore"; }
}
class Instruction_Astore_0 extends Instruction_noargs {
	Instruction_Astore_0() { super((byte)ByteCode.ASTORE_0); name = "astore_0"; }
}
class Instruction_Astore_1 extends Instruction_noargs {
	Instruction_Astore_1() { super((byte)ByteCode.ASTORE_1); name = "astore_1"; }
}
class Instruction_Astore_2 extends Instruction_noargs {
	Instruction_Astore_2() { super((byte)ByteCode.ASTORE_2); name = "astore_2"; }
}
class Instruction_Astore_3 extends Instruction_noargs {
	Instruction_Astore_3() { super((byte)ByteCode.ASTORE_3); name = "astore_3"; }
}
class Instruction_Iinc extends Instruction_bytevar {
	int arg_c;
	Instruction_Iinc() { super((byte)ByteCode.IINC); name = "iinc"; }
	public String toString() {
		return super.toString() + argsep + arg_c;
	}
	int nextOffset(int curr) { return super.nextOffset(curr)+1; }
	int parse(byte bc[],int index) { 
		index = super.parse(bc,index);
		arg_c = bc[index];
		return index+1; 
	}
	int compile(byte bc[],int index) { 
		index = super.compile(bc,index);
		bc[index] = (byte)(arg_c&0xff);
		return index+1; 
	}
}
class Instruction_Wide extends Instruction_bytevar {
	Instruction_Wide() { super((byte)ByteCode.WIDE); name = "wide"; }
}
class Instruction_Newarray extends Instruction {
	public static final int T_BOOLEAN = 4;
	public static final int T_CHAR = 5;
	public static final int T_FLOAT = 6;
	public static final int T_DOUBLE = 7;
	public static final int T_BYTE = 8;
	public static final int T_SHORT = 9;
	public static final int T_INT = 10;
	public static final int T_LONG = 11;
	byte atype;
	Instruction_Newarray() { super((byte)ByteCode.NEWARRAY); name = "newarray"; }
	public String toString() {
		String args;
		switch((int)atype) {
			case T_BOOLEAN: args = "boolean"; break;
			case T_CHAR: args = "char"; break;
			case T_FLOAT: args = "float"; break;
			case T_DOUBLE: args = "double"; break;
			case T_BYTE: args = "byte"; break;
			case T_SHORT: args = "short"; break;
			case T_INT: args = "int"; break;
			case T_LONG: args = "long"; break;
			default: args = Integer.toString(atype); break;
		}
		return super.toString() + argsep + args;
	}
	int nextOffset(int curr) { return curr+2; }
	int parse(byte bc[],int index) { 
		atype = bc[index];
		return index+1; 
	}
	int compile(byte bc[],int index) { 
		bc[index++] = code;
		bc[index] = atype;
		return index+1; 
	}
}
class Instruction_Anewarray extends Instruction_intindex {
	Instruction_Anewarray() { super((byte)ByteCode.ANEWARRAY); name = "anewarray"; }
}
class Instruction_Multinewarray extends Instruction_intindex {
	byte dims;
	Instruction_Multinewarray() { 
		super((byte)ByteCode.MULTINEWARRAY); 
		name = "multinewarray"; 
	}
	public String toString() {
		return super.toString() + argsep + dims;
	}
	int nextOffset(int curr) { return super.nextOffset(curr)+1; }
	int parse(byte bc[],int index) { 
		index = super.parse(bc,index);
		dims = bc[index];
		return index+1; 
	}
	int compile(byte bc[],int index) { 
		index = super.compile(bc,index);
		bc[index] = dims;
		return index+1; 
	}
}
class Instruction_Arraylength extends Instruction_noargs {
	Instruction_Arraylength() { super((byte)ByteCode.ARRAYLENGTH); name = "arraylength"; }
}
class Instruction_Iaload extends Instruction_noargs {
	Instruction_Iaload() { super((byte)ByteCode.IALOAD); name = "iaload"; }
}
class Instruction_Laload extends Instruction_noargs {
	Instruction_Laload() { super((byte)ByteCode.LALOAD); name = "laload"; }
}
class Instruction_Faload extends Instruction_noargs {
	Instruction_Faload() { super((byte)ByteCode.FALOAD); name = "faload"; }
}
class Instruction_Daload extends Instruction_noargs {
	Instruction_Daload() { super((byte)ByteCode.DALOAD); name = "daload"; }
}
class Instruction_Aaload extends Instruction_noargs {
	Instruction_Aaload() { super((byte)ByteCode.AALOAD); name = "aaload"; }
}
class Instruction_Baload extends Instruction_noargs {
	Instruction_Baload() { super((byte)ByteCode.BALOAD); name = "baload"; }
}
class Instruction_Caload extends Instruction_noargs {
	Instruction_Caload() { super((byte)ByteCode.CALOAD); name = "caload"; }
}
class Instruction_Saload extends Instruction_noargs {
	Instruction_Saload() { super((byte)ByteCode.SALOAD); name = "saload"; }
}
class Instruction_Iastore extends Instruction_noargs {
	Instruction_Iastore() { super((byte)ByteCode.IASTORE); name = "iastore"; }
}
class Instruction_Lastore extends Instruction_noargs {
	Instruction_Lastore() { super((byte)ByteCode.LASTORE); name = "lastore"; }
}
class Instruction_Fastore extends Instruction_noargs {
	Instruction_Fastore() { super((byte)ByteCode.FASTORE); name = "fastore"; }
}
class Instruction_Dastore extends Instruction_noargs {
	Instruction_Dastore() { super((byte)ByteCode.DASTORE); name = "dastore"; }
}
class Instruction_Aastore extends Instruction_noargs {
	Instruction_Aastore() { super((byte)ByteCode.AASTORE); name = "aastore"; }
}
class Instruction_Bastore extends Instruction_noargs {
	Instruction_Bastore() { super((byte)ByteCode.BASTORE); name = "bastore"; }
}
class Instruction_Castore extends Instruction_noargs {
	Instruction_Castore() { super((byte)ByteCode.CASTORE); name = "castore"; }
}
class Instruction_Sastore extends Instruction_noargs {
	Instruction_Sastore() { super((byte)ByteCode.SASTORE); name = "sastore"; }
}
class Instruction_Nop extends Instruction_noargs {
	Instruction_Nop() { super((byte)ByteCode.NOP); name = "nop"; }
}
class Instruction_Pop extends Instruction_noargs {
	Instruction_Pop() { super((byte)ByteCode.POP); name = "pop"; }
}
class Instruction_Pop2 extends Instruction_noargs {
	Instruction_Pop2() { super((byte)ByteCode.POP2); name = "pop2"; }
}
class Instruction_Dup extends Instruction_noargs {
	Instruction_Dup() { super((byte)ByteCode.DUP); name = "dup"; }
}
class Instruction_Dup2 extends Instruction_noargs {
	Instruction_Dup2() { super((byte)ByteCode.DUP2); name = "dup2"; }
}
class Instruction_Dup_x1 extends Instruction_noargs {
	Instruction_Dup_x1() { super((byte)ByteCode.DUP_X1); name = "dup_x1"; }
}
class Instruction_Dup2_x1 extends Instruction_noargs {
	Instruction_Dup2_x1() { super((byte)ByteCode.DUP2_X1); name = "dup2_x1"; }
}
class Instruction_Dup_x2 extends Instruction_noargs {
	Instruction_Dup_x2() { super((byte)ByteCode.DUP_X2); name = "dup_x2"; }
}
class Instruction_Dup2_x2 extends Instruction_noargs {
	Instruction_Dup2_x2() { super((byte)ByteCode.DUP2_X2); name = "dup2_x2"; }
}
class Instruction_Swap extends Instruction_noargs {
	Instruction_Swap() { super((byte)ByteCode.SWAP); name = "swap"; }
}
class Instruction_Iadd extends Instruction_noargs {
	Instruction_Iadd() { super((byte)ByteCode.IADD); name = "iadd"; }
}
class Instruction_Ladd extends Instruction_noargs {
	Instruction_Ladd() { super((byte)ByteCode.LADD); name = "ladd"; }
}
class Instruction_Fadd extends Instruction_noargs {
	Instruction_Fadd() { super((byte)ByteCode.FADD); name = "fadd"; }
}
class Instruction_Dadd extends Instruction_noargs {
	Instruction_Dadd() { super((byte)ByteCode.DADD); name = "dadd"; }
}
class Instruction_Isub extends Instruction_noargs {
	Instruction_Isub() { super((byte)ByteCode.ISUB); name = "isub"; }
}
class Instruction_Lsub extends Instruction_noargs {
	Instruction_Lsub() { super((byte)ByteCode.LSUB); name = "lsub"; }
}
class Instruction_Fsub extends Instruction_noargs {
	Instruction_Fsub() { super((byte)ByteCode.FSUB); name = "fsub"; }
}
class Instruction_Dsub extends Instruction_noargs {
	Instruction_Dsub() { super((byte)ByteCode.DSUB); name = "dsub"; }
}
class Instruction_Imul extends Instruction_noargs {
	Instruction_Imul() { super((byte)ByteCode.IMUL); name = "imul"; }
}
class Instruction_Lmul extends Instruction_noargs {
	Instruction_Lmul() { super((byte)ByteCode.LMUL); name = "lmul"; }
}
class Instruction_Fmul extends Instruction_noargs {
	Instruction_Fmul() { super((byte)ByteCode.FMUL); name = "fmul"; }
}
class Instruction_Dmul extends Instruction_noargs {
	Instruction_Dmul() { super((byte)ByteCode.DMUL); name = "dmul"; }
}
class Instruction_Idiv extends Instruction_noargs {
	Instruction_Idiv() { super((byte)ByteCode.IDIV); name = "idiv"; }
}
class Instruction_Ldiv extends Instruction_noargs {
	Instruction_Ldiv() { super((byte)ByteCode.LDIV); name = "ldiv"; }
}
class Instruction_Fdiv extends Instruction_noargs {
	Instruction_Fdiv() { super((byte)ByteCode.FDIV); name = "fdiv"; }
}
class Instruction_Ddiv extends Instruction_noargs {
	Instruction_Ddiv() { super((byte)ByteCode.DDIV); name = "ddiv"; }
}
class Instruction_Irem extends Instruction_noargs {
	Instruction_Irem() { super((byte)ByteCode.IREM); name = "irem"; }
}
class Instruction_Lrem extends Instruction_noargs {
	Instruction_Lrem() { super((byte)ByteCode.LREM); name = "lrem"; }
}
class Instruction_Frem extends Instruction_noargs {
	Instruction_Frem() { super((byte)ByteCode.FREM); name = "frem"; }
}
class Instruction_Drem extends Instruction_noargs {
	Instruction_Drem() { super((byte)ByteCode.DREM); name = "drem"; }
}
class Instruction_Ineg extends Instruction_noargs {
	Instruction_Ineg() { super((byte)ByteCode.INEG); name = "ineg"; }
}
class Instruction_Lneg extends Instruction_noargs {
	Instruction_Lneg() { super((byte)ByteCode.LNEG); name = "lneg"; }
}
class Instruction_Fneg extends Instruction_noargs {
	Instruction_Fneg() { super((byte)ByteCode.FNEG); name = "fneg"; }
}
class Instruction_Dneg extends Instruction_noargs {
	Instruction_Dneg() { super((byte)ByteCode.DNEG); name = "dneg"; }
}
class Instruction_Ishl extends Instruction_noargs {
	Instruction_Ishl() { super((byte)ByteCode.ISHL); name = "ishl"; }
}
class Instruction_Ishr extends Instruction_noargs {
	Instruction_Ishr() { super((byte)ByteCode.ISHR); name = "ishr"; }
}
class Instruction_Iushr extends Instruction_noargs {
	Instruction_Iushr() { super((byte)ByteCode.IUSHR); name = "iushr"; }
}
class Instruction_Lshl extends Instruction_noargs {
	Instruction_Lshl() { super((byte)ByteCode.LSHL); name = "lshl"; }
}
class Instruction_Lshr extends Instruction_noargs {
	Instruction_Lshr() { super((byte)ByteCode.LSHR); name = "lshr"; }
}
class Instruction_Lushr extends Instruction_noargs {
	Instruction_Lushr() { super((byte)ByteCode.LUSHR); name = "lushr"; }
}
class Instruction_Iand extends Instruction_noargs {
	Instruction_Iand() { super((byte)ByteCode.IAND); name = "iand"; }
}
class Instruction_Land extends Instruction_noargs {
	Instruction_Land() { super((byte)ByteCode.LAND); name = "land"; }
}
class Instruction_Ior extends Instruction_noargs {
	Instruction_Ior() { super((byte)ByteCode.IOR); name = "ior"; }
}
class Instruction_Lor extends Instruction_noargs {
	Instruction_Lor() { super((byte)ByteCode.LOR); name = "lor"; }
}
class Instruction_Ixor extends Instruction_noargs {
	Instruction_Ixor() { super((byte)ByteCode.IXOR); name = "ixor"; }
}
class Instruction_Lxor extends Instruction_noargs {
	Instruction_Lxor() { super((byte)ByteCode.LXOR); name = "lxor"; }
}
class Instruction_I2l extends Instruction_noargs {
	Instruction_I2l() { super((byte)ByteCode.I2L); name = "i2l"; }
}
class Instruction_I2f extends Instruction_noargs {
	Instruction_I2f() { super((byte)ByteCode.I2F); name = "i2f"; }
}
class Instruction_I2d extends Instruction_noargs {
	Instruction_I2d() { super((byte)ByteCode.I2D); name = "i2d"; }
}
class Instruction_L2i extends Instruction_noargs {
	Instruction_L2i() { super((byte)ByteCode.L2I); name = "l2i"; }
}
class Instruction_L2f extends Instruction_noargs {
	Instruction_L2f() { super((byte)ByteCode.L2F); name = "l2f"; }
}
class Instruction_L2d extends Instruction_noargs {
	Instruction_L2d() { super((byte)ByteCode.L2D); name = "l2d"; }
}
class Instruction_F2i extends Instruction_noargs {
	Instruction_F2i() { super((byte)ByteCode.F2I); name = "f2i"; }
}
class Instruction_F2l extends Instruction_noargs {
	Instruction_F2l() { super((byte)ByteCode.F2L); name = "f2l"; }
}
class Instruction_F2d extends Instruction_noargs {
	Instruction_F2d() { super((byte)ByteCode.F2D); name = "f2d"; }
}
class Instruction_D2i extends Instruction_noargs {
	Instruction_D2i() { super((byte)ByteCode.D2I); name = "d2i"; }
}
class Instruction_D2l extends Instruction_noargs {
	Instruction_D2l() { super((byte)ByteCode.D2L); name = "d2l"; }
}
class Instruction_D2f extends Instruction_noargs {
	Instruction_D2f() { super((byte)ByteCode.D2F); name = "d2f"; }
}
class Instruction_Int2byte extends Instruction_noargs {
	Instruction_Int2byte() { super((byte)ByteCode.INT2BYTE); name = "int2byte"; }
}
class Instruction_Int2char extends Instruction_noargs {
	Instruction_Int2char() { super((byte)ByteCode.INT2CHAR); name = "int2char"; }
}
class Instruction_Int2short extends Instruction_noargs {
	Instruction_Int2short() { super((byte)ByteCode.INT2SHORT); name = "int2short"; }
}
class Instruction_Ifeq extends Instruction_intbranch {
	Instruction_Ifeq() { super((byte)ByteCode.IFEQ); name = "ifeq"; }
}
class Instruction_Ifnull extends Instruction_intbranch {
	Instruction_Ifnull() { super((byte)ByteCode.IFNULL); name = "ifnull"; }
}
class Instruction_Iflt extends Instruction_intbranch {
	Instruction_Iflt() { super((byte)ByteCode.IFLT); name = "iflt"; }
}
class Instruction_Ifle extends Instruction_intbranch {
	Instruction_Ifle() { super((byte)ByteCode.IFLE); name = "ifle"; }
}
class Instruction_Ifne extends Instruction_intbranch {
	Instruction_Ifne() { super((byte)ByteCode.IFNE); name = "ifne"; }
}
class Instruction_Ifgt extends Instruction_intbranch {
	Instruction_Ifgt() { super((byte)ByteCode.IFGT); name = "ifgt"; }
}
class Instruction_Ifge extends Instruction_intbranch {
	Instruction_Ifge() { super((byte)ByteCode.IFGE); name = "ifge"; }
}
class Instruction_Ifnonnull extends Instruction_intbranch {
	Instruction_Ifnonnull() { super((byte)ByteCode.IFNONNULL); name = "ifnonnull"; }
}
class Instruction_If_icmpeq extends Instruction_intbranch {
	Instruction_If_icmpeq() { super((byte)ByteCode.IF_ICMPEQ); name = "if_icmpeq"; }
}
class Instruction_If_icmplt extends Instruction_intbranch {
	Instruction_If_icmplt() { super((byte)ByteCode.IF_ICMPLT); name = "if_icmplt"; }
}
class Instruction_If_icmple extends Instruction_intbranch {
	Instruction_If_icmple() { super((byte)ByteCode.IF_ICMPLE); name = "if_icmple"; }
}
class Instruction_If_icmpne extends Instruction_intbranch {
	Instruction_If_icmpne() { super((byte)ByteCode.IF_ICMPNE); name = "if_icmpne"; }
}
class Instruction_If_icmpgt extends Instruction_intbranch {
	Instruction_If_icmpgt() { super((byte)ByteCode.IF_ICMPGT); name = "if_icmpgt"; }
}
class Instruction_If_icmpge extends Instruction_intbranch {
	Instruction_If_icmpge() { super((byte)ByteCode.IF_ICMPGE); name = "if_icmpge"; }
}
class Instruction_Lcmp extends Instruction_noargs {
	Instruction_Lcmp() { super((byte)ByteCode.LCMP); name = "lcmp"; }
}
class Instruction_Fcmpl extends Instruction_noargs {
	Instruction_Fcmpl() { super((byte)ByteCode.FCMPL); name = "fcmpl"; }
}
class Instruction_Fcmpg extends Instruction_noargs {
	Instruction_Fcmpg() { super((byte)ByteCode.FCMPG); name = "fcmpg"; }
}
class Instruction_Dcmpl extends Instruction_noargs {
	Instruction_Dcmpl() { super((byte)ByteCode.DCMPL); name = "dcmpl"; }
}
class Instruction_Dcmpg extends Instruction_noargs {
	Instruction_Dcmpg() { super((byte)ByteCode.DCMPG); name = "dcmpg"; }
}
class Instruction_If_acmpeq extends Instruction_intbranch {
	Instruction_If_acmpeq() { super((byte)ByteCode.IF_ACMPEQ); name = "if_acmpeq"; }
}
class Instruction_If_acmpne extends Instruction_intbranch {
	Instruction_If_acmpne() { super((byte)ByteCode.IF_ACMPNE); name = "if_acmpne"; }
}
class Instruction_Goto extends Instruction_intbranch {
	Instruction_Goto() { super((byte)ByteCode.GOTO); name = "goto"; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = target;
		return i;
	}
}
class Instruction_Goto_w extends Instruction_longbranch {
	Instruction_Goto_w() { super((byte)ByteCode.GOTO_W); name = "goto_w"; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = target;
		return i;
	}
}
class Instruction_Jsr extends Instruction_intbranch {
	Instruction_Jsr() { super((byte)ByteCode.JSR); name = "jsr"; calls = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = target;
		return i;
	}
}
class Instruction_Jsr_w extends Instruction_longbranch {
	Instruction_Jsr_w() { super((byte)ByteCode.JSR_W); name = "jsr_w"; calls = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = target;
		return i;
	}
}
class Instruction_Ret extends Instruction_bytevar {
	Instruction_Ret() { super((byte)ByteCode.RET); name = "ret"; branches = true; returns = true; }
}
class Instruction_Ret_w extends Instruction_intvar {
	Instruction_Ret_w() { super((byte)ByteCode.RET_W); name = "ret_w"; branches = true; returns = true; }
}
class Instruction_Return extends Instruction_noargs {
	Instruction_Return() { super((byte)ByteCode.RETURN); name = "return"; branches = true; returns = true; }
}
class Instruction_Ireturn extends Instruction_noargs {
	Instruction_Ireturn() { super((byte)ByteCode.IRETURN); name = "ireturn"; branches = true; returns = true; }
}
class Instruction_Lreturn extends Instruction_noargs {
	Instruction_Lreturn() { super((byte)ByteCode.LRETURN); name = "lreturn"; branches = true; returns = true; }
}
class Instruction_Freturn extends Instruction_noargs {
	Instruction_Freturn() { super((byte)ByteCode.FRETURN); name = "freturn"; branches = true; returns = true; }
}
class Instruction_Dreturn extends Instruction_noargs {
	Instruction_Dreturn() { super((byte)ByteCode.DRETURN); name = "dreturn"; branches = true; returns = true; }
}
class Instruction_Areturn extends Instruction_noargs {
	Instruction_Areturn() { super((byte)ByteCode.ARETURN); name = "areturn"; branches = true; returns = true; }
}
class Instruction_Breakpoint extends Instruction_noargs {
	Instruction_Breakpoint() { super((byte)ByteCode.BREAKPOINT); name = "breakpoint"; }
}
class Instruction_Tableswitch extends Instruction {
	Instruction_Tableswitch() { super((byte)ByteCode.TABLESWITCH); name = "tableswitch"; branches = true; }
	byte pad;  // number of bytes used for padding
	int default_offset;
	int low;
	int high;
	int jump_offsets[];
	Instruction default_inst;
	Instruction jump_insts[];
	public String toString() {
		String args;
		int i;
		args = super.toString() + argsep + "(" + Integer.toString(pad) + ")";
		args = args + argsep + "label_" + Integer.toString(default_inst.label);
		args = args + argsep + Integer.toString(low);
		args = args + argsep + Integer.toString(high) + ": ";
		for (i=0;i<high-low+1;i++) {
			args = args + argsep + "label_" + Integer.toString(jump_insts[i].label);
		}
		return args;
	}
	int parse(byte bc[],int index) { 
		// first figure out padding to next 4-byte quantity
		int i,j,baseindex;
		baseindex = index;
		i = index % 4;
		if (i != 0)
			pad = (byte)(4 - i);
		else
			pad = (byte)0;
		index += pad;
		default_offset = getInt(bc,index);
		index += 4;
		low = getInt(bc,index);
		index += 4;
		high = getInt(bc,index);
		index += 4;
		i = high-low+1;
		if (i>0) {	
			jump_offsets = new int[i];
			j = 0;
			do {
				jump_offsets[j] = getInt(bc,index);
				index += 4;
				j++;
			} while(j<i);
		}
		return index;
	}
	int nextOffset(int curr) { 
		int i,j,baseindex,siz=0;
		baseindex = curr;
		i = (curr+1) % 4;
		if (i != 0)
			siz = (4 - i);
		return (curr + siz + 13 + (high-low+1)*4);
	}
	int compile(byte bc[],int index) { 
		int i;
		bc[index++] = code;
		// insert padding so next instruction is on a 4-byte boundary
		for (i=0;i<pad;i++) 
			bc[index++] = 0;
		if (default_inst!=null)
			index = intToBytes(default_inst.label-label,bc,index); 
		else
			index = intToBytes(default_offset,bc,index); 
		index = intToBytes(low,bc,index); 
		index = intToBytes(high,bc,index); 
		for (i=0;i<=high-low;i++) {
			if (jump_insts[i]!=null)
				index = intToBytes((jump_insts[i]).label-label,bc,index);
			else
				index = intToBytes(jump_offsets[i],bc,index);
		}
		return index; 
	}
	void offsetToPointer(ByteCode bc) { 
		int i;
		default_inst = bc.locateInst(default_offset+label);
		if (default_inst==null) {
			System.out.println("Warning: can't locate target of instruction");
			System.out.println(" which should be at byte address " + (label+default_offset));
		} else
		  default_inst.labelled = true;
		if (high-low+1>0) {
			jump_insts = new Instruction[high-low+1];
			for (i=0;i<high-low+1;i++) {
				jump_insts[i] = bc.locateInst(jump_offsets[i]+label);
				if (jump_insts[i]==null) {
					System.out.println("Warning: can't locate target of instruction");
					System.out.println(" which should be at byte address " + 
					                   (label+jump_offsets[i]));
				} else
				  jump_insts[i].labelled = true;
			}
		}
	}
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[high-low+2];
		int j;
		i[0] = default_inst;
		for (j=1;j<high-low+2;j++)
		  i[j] = jump_insts[j-1];
		return i;
	}
}
class Instruction_Lookupswitch extends Instruction {
	Instruction_Lookupswitch() { super((byte)ByteCode.LOOKUPSWITCH); name = "lookupswitch"; branches = true; }
	byte pad;  // number of bytes used for padding
	int default_offset;
	int npairs;
	int match_offsets[];
	Instruction default_inst;
	Instruction match_insts[];
	public String toString() {
		// first figure out padding to next 4-byte quantity
		String args;
		int i;
		args = super.toString() + argsep + "(" + Integer.toString(pad) + ")";
		args = args + argsep + Integer.toString(default_inst.label);
		args = args + argsep + Integer.toString(npairs) + ": ";
		for (i=0;i<npairs;i++)
			args = args + "case " + Integer.toString(match_offsets[i*2]) +
			       ": label_" + Integer.toString(match_insts[i].label);
		return args;
	}
	int parse(byte bc[],int index) { 
		// first figure out padding to next 4-byte quantity
		int i,j,baseindex;
		baseindex = index;
		i = index % 4;
		if (i != 0)
			pad = (byte)(4 - i);
		else
			pad = (byte)0;
		index += pad;
		default_offset = getInt(bc,index);
		index += 4;
		npairs = getInt(bc,index);
		index += 4;
		if (npairs>0) {	
			match_offsets = new int[npairs*2];
			j = 0;
			do {
				match_offsets[j] = getInt(bc,index);
				j++;
				index += 4;
				match_offsets[j] = getInt(bc,index);
				index += 4;
				j++; 
			} while(j<npairs*2);
		}
		return index;
	}
	int nextOffset(int curr) { 
		int i,j,baseindex,siz=0;
		baseindex = curr;
		i = (curr+1) % 4;
		if (i != 0)
			siz = (4 - i);
		return (curr + siz + 9 + npairs*8);
	}
 	int compile(byte bc[],int index) { 
		int i;
		bc[index++] = code;
		// insert padding so next instruction is on a 4-byte boundary
		for (i=0;i<pad;i++) 
			bc[index++] = 0;
		if (default_inst!=null)
			index = intToBytes(default_inst.label-label,bc,index); 
		else
			index = intToBytes(default_offset,bc,index); 
		index = intToBytes(npairs,bc,index); 
		for (i=0;i<npairs;i++) {
			index = intToBytes(match_offsets[i*2],bc,index);
			if (match_insts[i]!=null)
				index = intToBytes((match_insts[i]).label-label,bc,index);
			else
				index = intToBytes(match_offsets[i*2+1],bc,index);
		}
		return index; 
	}
	void offsetToPointer(ByteCode bc) { 
		int i;
		default_inst = bc.locateInst(default_offset+label);
		if (default_inst==null) {
			System.out.println("Warning: can't locate target of instruction");
			System.out.println(" which should be at byte address " + (label+default_offset));
		} else
		  default_inst.labelled = true;
		if (npairs>0) {
			match_insts = new Instruction[npairs];
			for (i=0;i<npairs;i++) {
				match_insts[i] = bc.locateInst(match_offsets[i*2+1]+label);
				if (match_insts[i]==null) {
					System.out.println("Warning: can't locate target of instruction");
					System.out.println(" which should be at byte address " + 
					                   (label+match_offsets[i*2+1]));
				} else
				  match_insts[i].labelled = true;
			}
		}
	}
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[npairs+1];
		int j;
		i[0] = default_inst;
		for (j=1;j<npairs+1;j++)
		  i[j] = match_insts[j-1];
		return i;
	}
}
class Instruction_Putfield extends Instruction_intindex {
	Instruction_Putfield() { super((byte)ByteCode.PUTFIELD); name = "putfield"; }
}
class Instruction_Getfield extends Instruction_intindex {
	Instruction_Getfield() { super((byte)ByteCode.GETFIELD); name = "getfield"; }
}
class Instruction_Putstatic extends Instruction_intindex {
	Instruction_Putstatic() { super((byte)ByteCode.PUTSTATIC); name = "putstatic"; }
}
class Instruction_Getstatic extends Instruction_intindex {
	Instruction_Getstatic() { super((byte)ByteCode.GETSTATIC); name = "getstatic"; }
}
class Instruction_Invokevirtual extends Instruction_intindex {
	Instruction_Invokevirtual() { super((byte)ByteCode.INVOKEVIRTUAL); name = "invokevirtual"; branches = true; calls = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = null; 
		return i;
	}
}
class Instruction_Invokenonvirtual extends Instruction_intindex {
	Instruction_Invokenonvirtual() { super((byte)ByteCode.INVOKENONVIRTUAL); name = "invokenonvirtual"; branches = true; calls = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = null;
		return i;
	}
}
class Instruction_Invokestatic extends Instruction_intindex {
	Instruction_Invokestatic() { super((byte)ByteCode.INVOKESTATIC); name = "invokestatic"; branches = true; calls = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = null;
		return i;
	}
}
class Instruction_Invokeinterface extends Instruction_intindex {
	Instruction_Invokeinterface() { 
		super((byte)ByteCode.INVOKEINTERFACE); 
		name = "invokeinterface"; 
		branches = true; 
		calls = true;
	}
	byte nargs;
	byte reserved;
	public String toString() {
		return super.toString() + argsep + nargs + argsep + "(reserved " + reserved + ")";
	}
	int nextOffset(int curr) { return super.nextOffset(curr)+2; }
	int parse(byte bc[],int index) { 
		index = super.parse(bc,index);
		nargs = bc[index];
		index++;
		reserved = bc[index];
		return index+1; 
	}
 	int compile(byte bc[],int index) { 
		index = super.compile(bc,index);
		bc[index++] = nargs;
		bc[index++] = reserved;
		return index; 
	}
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = null;
		return i;
	}
}
class Instruction_Athrow extends Instruction_noargs {
	Instruction_Athrow() { super((byte)ByteCode.ATHROW); 
						   name = "athrow"; branches = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = null;
		return i;
	}
}
class Instruction_New extends Instruction_intindex {
	Instruction_New() { super((byte)ByteCode.NEW); 
						name = "new"; branches = true; calls = true; }
	Instruction[] branchpoints(Instruction next) { 
		Instruction i[] = new Instruction[1];
		i[0] = null;
		return i;
	}
}
class Instruction_Checkcast extends Instruction_intindex {
	Instruction_Checkcast() { super((byte)ByteCode.CHECKCAST); name = "checkcast"; }
}
class Instruction_Instanceof extends Instruction_intindex {
	Instruction_Instanceof() { super((byte)ByteCode.INSTANCEOF); name = "instanceof"; }
}
class Instruction_Monitorenter extends Instruction_noargs {
	Instruction_Monitorenter() { super((byte)ByteCode.MONITORENTER); name = "monitorenter"; }
}
class Instruction_Monitorexit extends Instruction_noargs {
	Instruction_Monitorexit() { super((byte)ByteCode.MONITOREXIT); name = "monitorexit"; }
}
